#!/usr/bin/perl -w

######################################################################
# Crawl script
# Kevin Beason
# 4/12/01
######################################################################

use strict;
use XML::Parser;
use Getopt::Std;

# YOU MUST SET THESE IN THIS FILE
my $crawler_dir     = "/home/beason/bin/crawl";
my $config_file     = "$crawler_dir/config.xml";

# These can be reconfigured in the config.xml
my $crawl_url        = "/~beason/crawl";
my $crawler_subdir   = ".crawler_files";
my $crawler_page     = ".crawler_index.html";
my $mirror_url       = "/~beason/mirrors";
my $mirror_dir       = "/rivera/a/beason/mirrors";
my $old_mirror_dir   = "/rivera/a/beason/old_mirrors";
my $types_file       = "$crawler_dir/types.xml";
my $html_shell       = "$crawler_dir/shell.html";
my $dir_list         = "$crawler_dir/dirs.txt";
my $ln_cmd           = "/sbin/ln";
my $drawshapes_cmd   = "/home/banks/bin/drawShapes";
my $convert_cmd      = "/usr/local/bin/convert";
my $mogrify_cmd      = "/usr/local/bin/mogrify";
my $dmconvert_cmd    = "/usr/sbin/dmconvert";
my $mv_cmd           = "/bin/mv";
my $cp_cmd           = "/bin/cp";
my $drawshapes_param = ".drawShapes";
my $thumbnail_width  = 100;
my $thumbnail_height = 100;
my $thumbnails_per_row = 4;

######################################################################
# Read Config File
######################################################################

read_config_file();
#test_read_config_file();

######################################################################
# Read File Types
######################################################################

my %filetypes;
read_file_types();
#test_read_file_types();

######################################################################
# Loop through students, and mirror their directory
######################################################################

my @mirror_these;

# if command line parameters, assume they are directories to crawl
if (@ARGV){
    @mirror_these = @ARGV;
}
# else clear mirror directory and crawl the dirs in dirs.txt
else{
    # Read directory list
    open DIRLIST, "<$dir_list";
    while (<DIRLIST>){
	chomp;
	next if /^#/;
	next unless /[^\s]/;
	push @mirror_these, $_;
    }
    close DIRLIST;

    # clear mirror directory
    my $time = time;
    system($mv_cmd, "$mirror_dir", "$old_mirror_dir/$time");
    if ($?){
	die "Could not move $mirror_dir: $?";
    }
    create_dir($mirror_dir);
    copy_file("$mirror_dir", "$crawler_dir/$crawler_subdir/.htaccess");
}


# build mirror

my $dir;
my $dir_index = "";
my $root;
my $target_dir;
foreach $dir (@mirror_these){
    $root = $dir;
    $root =~ s/^.*\///; # remove everything up to and including last slash
    $target_dir = "$mirror_dir/$root";

    create_dir($target_dir);
    create_dir("$target_dir/$crawler_subdir");
    makedrawfiles("$target_dir/$crawler_subdir", $dir);

    mirrorize($target_dir, $dir, 0);

    $dir_index .= "<br>\n" if ($dir_list);
    $dir_index .= "<a href=\"$root/$crawler_page\">$dir</a>";
}

# create root webpage

# construct %data structure for writepage()
my %data;
opendir(DIR, $mirror_dir) or die $!;
while ($_ = readdir DIR){
    next if /^\./;
    next unless (-d "$mirror_dir/$_");
    push @{ $data{dir} }, $_;
}
closedir DIR;

# create background images
create_dir("$mirror_dir/$crawler_subdir");
makedrawfiles("$mirror_dir/$crawler_subdir", $mirror_dir);

# create page
write_page($mirror_dir, $mirror_dir, \%data, 0);


######################################################################
# Exit
######################################################################

print "Done!\n";
exit;


######################################################################
# subroutine mirrorize - Mirrorize a dir
#     $mirror_target = the directory to create the webpage
#     $dir           = the directory having the content (student dir)
#     $depth         = level of recursion (first call depth is 0)
#
# $data{"filetype"} = [ "file1", "file2" ]
#
######################################################################

sub mirrorize {
    my ($mirror_target, $dir, $depth) = @_;

    # create page

    # examine list of files in directory, building %data
    my $filename;
    my $full_filename;
    my $extension;
    my %data;
    my %merged_data;
    
    opendir(DIR, $dir) or do {
	warn "Couldn't open directory $dir: $!";
	next;
    };
    
    while ($filename = readdir DIR){
	print STDERR "Looking at $dir/$filename\n";
	
	# skip . and ..
	next if $filename =~ /^(\.|\.\.)$/;

	# skip hidden files and directories
	next if $filename =~ /^\./;

	# skip the mirror directories!
	$_ = "$dir/$filename";
	next if (/^$mirror_dir/);
	next if (/^$old_mirror_dir/);

	$full_filename = "$dir/$filename";

	# skip undreadable files
	next unless (-r $full_filename);
	if (-l $full_filename){
	    next unless (-r readlink($full_filename));
	}

	# if directory, add to list of dirs
	if (-d $full_filename){
	    next if (-l $full_filename); # skip directory links
	    
	    push @{ $data{"dir"} }, $filename;
	    next;
	}

	print "Attempting to match a type to $filename (haven't skipped)\n";

	# match with file type based on name
	my $filetype;
	my $known_name;
	my $skip_ext_matching = 0;
	foreach $filetype (keys %filetypes){
	    foreach $known_name (@{ $filetypes{$filetype}->{names} }){
		$_ = lc $filename;
		$known_name = lc $known_name;
		if ($known_name eq $_){
		    # We have a match
		    print "Found $filetype file $_\n";
		    push @{ $data{$filetype} }, $filename;
		    if ($filetypes{$filetype}->{options}->{list} eq "no"){
			# don't bother matching an extension that
			# will not get listed. This also prevents it from
			# being listed, which is what we want.
			$skip_ext_matching = 1;
			last;
		    }
		}
	    }
	    # even if $skip_ext_matching is 1,
	    # check other file types incase file matches
	    # several file types
	}
	next if $skip_ext_matching;

	# match with file type based on extension
	# match extension with one in %filetypes

	print "Attempting to match a type (2) to $filename\n";
	
	# loop over every extension for every file type
	my $known_extension;
	my $break = 0;	
	foreach $filetype (keys %filetypes){
	    foreach $known_extension (@{ $filetypes{$filetype}->{extensions} }){
		$_ = quotemeta $known_extension; # escape the extension
		if ($filename =~ /$_$/i){
  		    push @{ $data{$filetype} }, $filename;
		    $break = $filetype;
		}
		last if ($break);
	    }
	    last if ($break);
	}
	print("Done matching filetype (it was $break)\n") if $break;
    }
    closedir DIR;

    if (%data){
	# mirrorize sub directories (recursive)
	my $subdir;
	
	# %merged_data = %data; # WRONG!!!!!
	foreach (keys %data){
	    @{ $merged_data{$_} } = @{ $data{$_} };
	}
	
	foreach $subdir (@{ $data{"dir"} }){
	    print "Mirrorizing(depth $depth) subdir $subdir\n";
	    print "BTW: \$data{dir} = \n    " .
		join("\n    ", @{ $data{dir} }) . "\n";
	    create_dir("$mirror_target/$subdir");
	    create_dir("$mirror_target/$subdir/$crawler_subdir");
	    my $subdir_data = mirrorize("$mirror_target/$subdir",
					"$dir/$subdir", $depth+1);
	    # if no files in that directory to list,
	    # or if only listed files are marked as "don't link",
	    # then don't link to them
	    my $filetype;
	    my $filename;
	    my $listed_files = 0;
	    foreach $filetype (keys %$subdir_data){
		next if $filetype =~ /dontlink/;  # not a proper filetype
		foreach $filename ( @{ $subdir_data->{$filetype} } ){
		    if (! $subdir_data->{"dontlink"}->{$filename}){
			$listed_files++;
			last;
		    }
		}
		last if $listed_files;
	    }

	    print "BTW2: \$data{dir} = \n    " .
		join("\n    ", @{ $data{dir} }) . "\n";

	    
	    if ($listed_files == 0){
		$data{"dontlink"}->{$subdir}++;
		$merged_data{"dontlink"}->{$subdir}++;
	    }
	    else{
		# check incorporate option
		print "Checking file types for $subdir for merge:";
		my $known_name;
		my $do_merge = 0;
		foreach $filetype (keys %filetypes){
		    print " $filetype";
		    foreach $known_name (@{ $filetypes{$filetype}->{names} }){
			$_ = lc $subdir;
			$known_name = lc $known_name;
			if ($known_name eq $_){
			    # We found a matching filetype
			    if ($filetypes{$filetype}->{options}->{incorporate}
				eq "yes"){
				print "Found merge for filetype $filetype\n";
				$do_merge = 1;
				last;
			    }
			}
		    }
		    last if $do_merge;
		}
		
		print "BTW3: \$data{dir} = \n    " .
		join("\n    ", @{ $data{dir} }) . "\n";

  	    	if ($do_merge){
		    print "Doing merge of subdir_data filetype:\n";
		    # incorporate the directory
		    # i.e., merge $subdir_data into $data
		    foreach $filetype (keys %$subdir_data){
			if ($filetype eq "dontlink"){
			    foreach (keys %{ $subdir_data->{dontlink} }){
				$merged_data{dontlink}->{"$subdir/$_"}++;
			    }
			}
			else{
			print "    $filetype:" .
			    join(", ",@{$subdir_data->{$filetype}}) . "\n";
			    foreach $filename (@{ $subdir_data->{$filetype} }){
				print "Merging file $subdir/$filename\n";
				unshift @{ $merged_data{$filetype} },
			            "$subdir/$filename";
			    }
			}
		    }
		}
		# end incorporate option check

		print "BTW4: \$data{dir} = \n    " .
		join("\n    ", @{ $data{dir} }) . "\n";

	    }
	}

	print "BTW5: \$data{dir} = \n    " .
	    join("\n    ", @{ $data{dir} }) . "\n";

	# make the page
	write_page($mirror_target, $dir, \%merged_data, $depth);
	
	print "BTW6: \$data{dir} = \n    " .
	    join("\n    ", @{ $data{dir} }) . "\n";
    }
	
    return \%data;
}

######################################################################
# subroutine write_page - Writes HTML page
#     $target_dir = the target directory to write the index.html file
#     $source_dir = the directory the filenames supplied in the
#                   next argument are relative to
#     $p = pointer to data structure needed to create the html page
#     $depth = level of recursion in mirrorize()
######################################################################

sub write_page {
    my ($target_dir, $source_dir, $p, $depth) = @_;  # p = data pointer

    # read prototype shell
    open(SHELL, "<$html_shell") or die "$!";
    my $shell;
    $shell = join "", <SHELL>;
    close SHELL;
    
    # open file
    open(PAGE, ">$target_dir/$crawler_page") or
	die "Could not open $target_dir/$crawler_page for writing: $!";

    # write file

    # start with directories, then loop through file types
    my $indexes = "";
    my $contact_info;
    my $main_content = "";
    my $type;
    my $file;
    my $title = "";
    my $body = "";

    # sort types, but put dir first
    my %types_list;
    my @types_list;
    foreach (keys %$p){
	$types_list{$_}++;
    }
    if ($types_list{dir}){
	delete $types_list{dir};
	push @types_list, "dir";
    }
    push @types_list, sort(keys %types_list);
    
    foreach $type (@types_list){
	next if ($type eq "dontlink");  # dontlink isn't a file type but a
	                                # helper group, indicating empty
	                                # or undreadable files.
	next unless (@{ $p->{$type} }); # skip type unless there are files
	                                # OF this type

	# set any variables
	if ($filetypes{$type}->{options}->{setvar}){
	    foreach $file ( @{ $p->{$type} } ){
		# read contents of file
		print "Reading contents of $file\n";
		my $file_contents;
		open FILE, "<$source_dir/$file" or die $!;
		$file_contents = join "", <FILE>;
		close FILE;

		# set the variable
		$_ = "\$" . $filetypes{$type}->{options}->{setvar} .
		     "=\$file_contents";
		print "Evaluating $_\n";
		eval $_;
	    }
	}
	
	# dont list this type if "list" option is "no" in types file
	next if ($filetypes{$type}->{options}->{list} eq "no");
	
	my $dir;
	my $link;
	my $thumb_num = 0;
        my $thumbnail = 0;
	my $caption = 0;
        if ($filetypes{$type}->{options}->{thumbnail} eq "yes"){
	    $thumbnail = 1;
	}
	if ($filetypes{$type}->{options}->{caption} eq "yes"){
	    $caption = 1;
	}

	# do caption processing
	my %captionlist;
	if ($caption){
	    foreach $file (@{ $p->{$type} }){ # for each file
		print "Searching for caption for file $file\n";
		my $caption_text = "";
		# search for caption file
		my $file_root;
		my $text_file;
		foreach $text_file ( @{ $p->{textfile} } ){
		    $file_root = $file;
		    $file_root =~ s/\.[^.]*?$// or next; # remove extension
		    if ($text_file =~ /^$file_root(\.txt|\.TXT)$/){
                        print "Match found!\n";
			open(CAPTION, "<$source_dir/$text_file")
			    or do {print "Unable to open file $!"; last};
			$caption_text = join "", <CAPTION>;
			close CAPTION;
			last;
		    }
		}
		if ($caption_text){
		    $captionlist{$file} = $caption_text;
		}
	    }
	}
	# if $type="textfile", don't list captions
	my %is_a_caption;
	if ($type eq "textfile"){
	    my $type;
	    foreach $type (keys %$p){
		next if ($type eq "dontlink"); # special type
		next unless ($filetypes{$type}->{options}->{caption} eq "yes");
		foreach $file (@{ $p->{$type} }){ # for each file
		    my $file_root;
		    my $text_file;
		    foreach $text_file ( @{ $p->{textfile} } ){
			$file_root = $file;
			$file_root =~ s/\.[^.]*?$// or next;
			if ($text_file =~ /^$file_root(\.txt|\.TXT)$/){
			    $is_a_caption{$text_file}++;
			    last;
			}
		    }
		}
	    }
	    # skip textfile type altogether if all they are are captions
	    if (@{ $p->{textfile} } == keys(%is_a_caption)){
		next;
	    }
	}
	
	# do thumbnail processing
	# weed out files that already have thumbnails,
	# thumbnail the ones that don't
	my %filelist;
	my %thumblist;

	if ($thumbnail){
	    foreach $file (@{ $p->{$type} }){
		$filelist{$file} = "";
	    }
	    my $other_file;
	    my $file_path;
	    my $file_root;
	    my $file_ext;
	    my $target_name;
	    my $target_ext;

	    # Step 1: Associate thumbnails with what they thumbnail
	    foreach $file (keys %filelist){
		next unless (defined $filelist{$file});

		# search for a match
		foreach $other_file (keys %filelist){
		    next if $file eq $other_file;
		    $file_root = $file;
		    if ($file_root =~ s/^(.*\/)//){
			$file_path = $1;
		    }
		    else { $file_path = "" }
		    if ($file_root =~ s/(\.[^.]*)$//){
			$file_ext = $1;
		    }
		    else {
			$file_ext = "";
			# no way
			next;
		    }
		    my $ext_re = "($file_ext|.jpg|.JPG|.jpeg|.JPEG|.gif|.GIF)";
		    my $sep_re = "[-_.]";
		    my $thm_re = "(t|T|th|TH|thumb|THUMB|tn|TN|" .
			         "sm|SM|small|SMALL)";

		    if ( # "pics/th_screen.jpg"
			$other_file =~
			/^$file_path$thm_re$sep_re$file_root$ext_re$/ ||

			# "pics/screen_tn.jpg"
			$other_file =~
			/^$file_path$file_root$sep_re$ext_re$/){

			# match found
			# associate with file, eliminate it from viewing
			$filelist{$file} = $other_file;
			delete $filelist{$other_file};
			
		    }
		}
		
	    }

	    # Step 2: Create thumbnails
	    foreach $file (keys %filelist){
		
		# Step 2a: determine appropriate name for thumbnail
		$file_root = $file;
		if ($file_root =~ s/^(.*\/)//){
		    $file_path = $1; # note that this includes a trailing "/"
		}
		else { $file_path = "" }
		if ($file_root =~ s/(\.[^.]*)$//){
		    $file_ext = $1;
		}
		else { $file_ext = "" }
		if ($file_ext =~ /^\.gif$/i){
		    $target_ext = ".gif";
		}
		else { $target_ext = ".jpg" }
		$target_name = "$file_path$crawler_subdir/$file_root$file_ext$target_ext";
		
		# Step 2b: Does thumbnail already exist?
		if (-e "$target_dir/$target_name"){
		    $thumblist{$file} = $target_name;
		    next;
		}

		# Step 2c: Does thumbnail file exist with another name?
		if ($filelist{$file}){
		    $other_file = $filelist{$file};

		    # Is it a jpeg or gif already?
		    if ($other_file =~ /(.gif|.GIF|.jpg|.JPG|.jpeg|.JPEG)$/){
			mirror_file("$target_dir/$crawler_subdir",
				    "$source_dir/$other_file",
				    "$target_dir/$target_name");
			$thumblist{$file} = $target_name;
			next;
		    }
		    else { # if not, convert it
			if ($type eq "image"){
			    system($convert_cmd,
				   "$source_dir/$other_file",
				   "$target_dir/$target_name");
			    if (!$?){ # if successful
				$thumblist{$file} = $target_name;
				next;
			    }
			}
			elsif ($type eq "movie"){
			    system($dmconvert_cmd, "-f", "rgb", "-p", "video",
				   $other_file,
				   "$target_dir/$target_name.rgb");
			    if (!$?){ # if successful
				system($convert_cmd,
				       "$target_dir/$target_name.rgb",
				       "$target_dir/$target_name");
				if (!$?){ # if successful
				    unlink "$target_dir/$target_name.rgb" or
					die $!;
				    $thumblist{$file} = $target_name;
				    next;
				}
			    }
			    # something went wrong
			    # continue to Step 2d
			}
		    }
		}
		# end Step 2c (thumbnail exists in source dir

		# Step 2d: Create new thumbnail
		$_ = thumbnail("$target_dir/$file_path$crawler_subdir",
			       "$source_dir/$file", $type);
		if ($_){
		    $thumblist{$file} = "$file_path$crawler_subdir/$_";
		    next;
		}
		else {
		    # something is wrong.. oh well, no thumbnail
		    next;
		}
	    }
	}
	# if image has a thumbnail, $thumblist{$file} has the value
	# if image IS a thumbnail, there is no $filelist{$file} defined

	# check for pluralName
	if (not defined $filetypes{$type}->{pluralName}){
	    warn "\$filetypes{\$type}->{pluralName} is not defined for" .
		" type $type";
	}

	# add indexes link for this type
	$indexes .= "<a href=\"#$type\">" . $filetypes{$type}->{pluralName} .
	    "</a><br>";

	# add heading for this type
	$main_content .= "<a name=\"$type\"><h3>" .
	    $filetypes{$type}->{pluralName} . "</h3></a>\n" .
	    "Click on something below.\n\n";

	# begin main entry for this type
	$main_content .= $thumbnail?"<table>\n":"<ul>\n";
	
	foreach $file (sort @{ $p->{$type} }){
	    next if ($thumbnail && !defined($filelist{$file})); # skip thumbs
	    next if ($type eq "textfile" && $is_a_caption{$file});

	    # list file
	    if ($type eq "dir"){ $dir = 1 }
	    else { $dir = 0 }
	    if ($p->{"dontlink"}->{$file}){ $link = 0 }
	    else { $link = 1 }

	    # I know this is unreadable. What this does though,
	    # is create an entry for a file, with all the dressings,
	    # depending on if it should be linked, tabled,
	    # is a directory, an image thumbnail, etc.
	    $_ = "";
	    $_ .= $thumbnail?"\n":"\n<li>";
	    $_ .= ($thumbnail&&$thumb_num%$thumbnails_per_row==0&&
		   $thumb_num!=0)?"</tr>":"";
	    $_ .= ($thumbnail&&$thumb_num%$thumbnails_per_row==0)?"<tr>\n":"";
	    $_ .= $thumbnail?"<td width=$thumbnail_width align=center" .
		  " valign=top>":"";
	    $_ .= $link?"<a href=\"$file".($dir?"/$crawler_page":""):"";
	    $_ .= $link?"\">":"";
	    $_ .= ($thumbnail&&$thumblist{$file}) ?
		"<img src=\"$thumblist{$file}\"><br>" :
		"";
	    $_ .= $file;
	    $_ .= $dir?"/":"";
            $_ .= $link?"</a>":"";
	    $_ .= ($type ne "dir")?" (" . filesize("$source_dir/$file") .
		  ")":"";
	    $_ .= ($caption&&$captionlist{$file})?
		"<br><font face=\"Arial, Helvetica\" size=1>" .
		$captionlist{$file} . "</font>":"";
	    $_ .= $thumbnail?"</td>":"";
	    $_ .= "\n";
	    $main_content .= $_;

	    mirror_file($target_dir, "$source_dir/$file")
		unless ($type eq "dir");

	    # increment $thumb_num
	    $thumb_num++;
	}

	$main_content .= $thumbnail?"</tr></table>\n":"</ul>\n";
    }

    # figure out title(s)
    @_ = split "/", $source_dir;
    @_ = reverse @_;
    for ($_=1;$_<$depth+1;$_++){
	$_[$_] = "<a href=\"" . "../"x$_ . "$crawler_page\">" .
	    $_[$_] . "</a>";
    }
    @_ = reverse @_;
    my $title_html = join "/", @_;
    my $title_plain;
    if ($title){
	$title_plain = $title;
	$title_html .= "<p>\n$title";
    }
    else{
	$title_plain = $source_dir;
    }
    
    # determine user and contact info
    my $user = $source_dir;
    if ($user =~ m#^/home/([-\w]+)/# or $user =~ m#^/home/([-\w]+)$#){
	$user = $1;
	my $full_name = (getpwnam($user))[5];
	$contact_info = <<EOF;
$full_name<br>\n
Mail: <i><a href="mailto:$user\@csit.fsu.edu/">$user\@csit.fsu.edu</a></i><br>
Web: <i><a href="http://www.csit.fsu.edu/~$user/">http://www.csit.fsu.edu/~$user/</a></i><br>
EOF
    }
    else{
	$contact_info = <<EOF;
Unknown user.<br>
<a href="mailto:webmaster\@csit.fsu.edu">E-mail the webmaster</a>
for more information.
EOF
    }

    my $dir_last_mod_date = date((stat($source_dir))[9]);
    my $page_created_date = date();
    my $pics_dir = "../"x$depth . "$crawler_subdir";


    # mirror links
    my $mirror_links = 
	(($target_dir ne $mirror_dir)?
          "<a href=\"$mirror_url/\">Mirror&nbsp;Home</a>\n" .
	 "<br><a href=\"../" . (($depth)?$crawler_page:"") .
          "\">Previous&nbsp;Directory</a>\n" :
	  ""
	);
    
    # make substitutions
    $shell =~ s/&&title_plain&&/$title_plain/sgi;
    $shell =~ s/&&title_html&&/$title_html/sgi;
    $shell =~ s/&&body&&/$body/sgi;
    $shell =~ s/&&mirror_url&&/$mirror_url\//sgi;
    $shell =~ s/&&crawler_page&&/$crawler_page/sgi;
    $shell =~ s/&&mirror_links&&/$mirror_links/sgi;
    $shell =~ s/&&indexes&&/$indexes/sgi;
    $shell =~ s/&&main_content&&/$main_content/sgi;
    $shell =~ s/&&contact_info&&/$contact_info/sgi;
    $shell =~ s/&&dir_last_mod_date&&/$dir_last_mod_date/sgi;
    $shell =~ s/&&page_created_date&&/$page_created_date/sgi;
    $shell =~ s/&&pics_dir&&/$pics_dir/sgi;
    $shell =~ s/&&crawl_url&&/$crawl_url\//sgi;

    # write page
    print PAGE $shell;

    # close file
    close PAGE;

    # link to index.html if it's not going to clobber
    if (! -e "$target_dir/index.html"){
	chdir $target_dir;
	system($ln_cmd, "-s", "$target_dir/$crawler_page",
	       "index.html");
    }
}

######################################################################
# subroutine create_dir - Ensures a directory exists
######################################################################

sub create_dir {
    my $target_dir = shift;

    if (-e $target_dir){
	if (! -d $target_dir){
	    die "Could not create dir $target_dir: file exists";
	}
	else {
	    # dir already exists
	}
    }
    else {
	# create dir
	mkdir("$target_dir", 0755) or
	    die "Could not make directory $target_dir: $!";
    }
}

######################################################################
# subroutine mirror_file - Mirrors the original file someplace else
#                          for public viewing. Returns target filename
#                          if successful.
#     $target_dir - directory to copy file to
#     $source_file - file to mirror
#     $target_file_name - OPTIONAL target file name
######################################################################

sub mirror_file {
    my ($target_dir, $source_file, $target_file_name) = @_;

    print "Mirroring $source_file to $target_dir...\n";
    
    my $target_file = $source_file;
    $target_file =~ s/^.*\///; # truncate directory
    $target_file = "$target_dir/$target_file";
    if ($target_file_name){
	$target_file = $target_file_name;
    }

    # make sure target_dir exists
    if (!(-e $target_dir && -d $target_dir)){
	warn "Could not copy file $source_file: \$target_dir $target_dir does not exist or is not a directory!";
	return;
    }

    # if owner of file is not owner of target_dir,
    # link will not work => copy the file
    if ( ((stat("$target_dir"))[4]) != ((stat("$source_file"))[4]) ){
	return copy_file($target_dir, $source_file, $target_file_name);
    }
    
    # link file to save space
    print STDERR "linking $target_file to $source_file\n";
    system($ln_cmd, "-s", "$source_file", "$target_file");
    if ($?){
	warn "Could not mirror file $source_file: Error($?) !";
	return;
    }
    
    return $target_file;    
}

######################################################################
# subroutine copy_files - Copy files. Returns target filename if
#                         successful.
#     $target_dir - directory to copy file to
#     $source_file - file to mirror
#     $target_file_name - OPTIONAL target file name
######################################################################

sub copy_file {
    my ($target_dir, $source_file, $target_file_name) = @_;
    my $target_file = $source_file;
    $target_file =~ s/^.*\///; # truncate directory
    $target_file = "$target_dir/$target_file";
    if ($target_file_name){
	$target_file = $target_file_name;
    }

    # make sure target_dir exists
    if (!(-e $target_dir && -d $target_dir)){
	warn "Could not copy file $source_file: \$target_dir $target_dir does not exist or is not a directory!";
	return;
    }

    # copy
    print STDERR "copying $source_file to $target_dir\n";
    system($cp_cmd, "$source_file", "$target_file");
    if ($?){
	warn "Could not copy file $source_file: Error($?) !";
	return;
    }
    chmod(0644, $target_file) or warn "Could not chmod file $target_file: $!";

    return $target_file;
}

######################################################################
# subroutine thumbnail - Thumbnail a file, returns thumbnail name
#                        sans path
#     $target_dir - place to put thumbnail
#     $file - file to thumbnail
#     $type - type of the file (key to %filetypes)
######################################################################

sub thumbnail {
    my ($target_dir, $file, $type) = @_;

    print "Thumbnailing $file...\n";

    # parse file name
    my $filename = $file;
    my $old_ext;
    $filename =~ s/^.*\///; # remove path prefix
    if ($filename =~ /(\.[^.]*?)$/){ # grab extension
	$old_ext = $1;
    }
    else{
	return;
    }

    if ($type eq "image"){
	# convert image to tga, resize, convert to jpg unless
	# it was originally .gif
	my $new_ext = ".jpg";
	if ($old_ext =~ /\.gif/i){
	    $new_ext = ".gif";
	}
	my $intermediate_img = "$target_dir/$filename.tga";
    
	system($convert_cmd, $file, $intermediate_img);
	if ($?){
	    warn "Error converting image $file to $intermediate_img: $?";
	    return;
	}

	# resize image
	system($mogrify_cmd, "-geometry",
	       "${thumbnail_width}x${thumbnail_height}>", $intermediate_img);
	if ($?){
	    warn "Error resizing image $intermediate_img: $?";
	    return;
	}

	# convert it to web format
	my $thumb_img = "$target_dir/$filename$new_ext";
	system($convert_cmd, $intermediate_img, $thumb_img);
	if ($?){
	    warn "Error converting image $intermediate_img to $thumb_img: $?";
	    return;
	}

	# clean up
	unlink($intermediate_img) or die "$!";
    
	return "$filename$new_ext";
    }
    elsif ($type eq "movie"){
	my $intermediate_img = "$target_dir/$filename.rgb";
	# grab first frame
	system($dmconvert_cmd, "-f", "rgb", "-p", "video",
	       $file, $intermediate_img);
	if ($?){
	    warn "Could not thumbnail movie $file: $?";
	    return;
	}

	# resize first frame
	system($mogrify_cmd, "-geometry",
	       "${thumbnail_width}x${thumbnail_height}>", $intermediate_img);
	if ($?){
	    warn "Error resizing image $intermediate_img: $?";
	    return;
	}

	# convert it to web format
	my $new_ext = ".jpg";
	my $thumb_img = "$target_dir/$filename$new_ext";
	if (-e $thumb_img){
	    warn "Movie thumbnail $thumb_img already exists? Not replacing.";
	}
	else {
	    system($convert_cmd, $intermediate_img, $thumb_img);
	    if ($?){
		warn "Error converting image $intermediate_img to " .
		    "$thumb_img: $?";
		return;
	    }
	}
	return "$filename$new_ext";
    }
}

######################################################################
# subroutine read_file_types -  Read file types file (it's in XML)
#
# XML tree diagram
#
# tree
#   |
#   |- "filetype_config"
#   `-.
#     |- filetype_config attributes
#     |
#     |- "filetype"
#     |-.
#     | |- filetype attributes
#     | |
#     | |- "ext"
#     | |-.
#     | | |- ext attributes
#     | | |
#     | | |- 0
#     | | |- ".txt"
#     | | 
#     | |- "ext"
#       ...
#
#
#   %filetypes{"file type"} = { pluralName => "name",
#                               extensions => [ "ext1", "ext2" ] }
#
######################################################################

sub read_file_types {

    my $parser = new XML::Parser(Style=>'Tree', ErrorContext=>2);
    my $tree = $parser->parsefile("$types_file");
    die '<filetype_config> isn\'t the top-level element'
	unless $tree->[0] eq "filetype_config";
    
    # read filetypes
    my $p = $tree->[1];  # p points to contents of filetype_config element

    # loop through contents of filetype_config
    for (my $i = 1; $i<@$p; $i+=2){

	# should not be seeing non-whitespace text, only elements
	if (not $p->[$i]){
	    warn 'text not allowed here' unless $p->[$i+1] =~ /^\s*$/;
	    next;
	}
	if ($p->[$i] ne "filetype"){
	    warn "<$p->[$i]> not allowed in <filetype_config>";
	    next;
	}

	my $p2 = $p->[$i+1];  # p2 points to contents of filetype element
	my $p3 = $p2->[0];    # p3 points to attribute hash of filetype element
	
	# get the name of the filetype
	my $typename = $p3->{'name'};

	# create data structure
	$filetypes{$typename} = {};
	$filetypes{$typename}->{ options } = { # options are special
	    "thumbnail"   => "no",             # parameters for processing  
	    "list"        => "yes",            # flags rather than data
	    "incorporate" => "no",
	    "caption"     => "no"
	    };
	$filetypes{$typename}->{ extensions } = [];
	$filetypes{$typename}->{ names      } = [];
	$filetypes{$typename}->{ pluralName } = "";
	$filetypes{$typename}->{ fullName   } = "";

	# loop through contents of filetype element
	for (my $j = 1; $j < @$p2; $j += 2){
	    if (not $p2->[$j]){
		warn "text not allowed in <filetype>" unless
		    $p2->[$j+1] =~ /^\s*$/;
		next;
	    }
	    elsif ($p2->[$j] eq "like"){
		# do nothing, for right now
	    }
	    elsif ($p2->[$j] eq "pluralName"){
		$p3 = $p2->[$j+1];
		if ($p3->[1]){
		    warn "<pluralName> can only contain text";
		}
		$filetypes{$typename}->{pluralName} = $p3->[2];
	    }
	    elsif ($p2->[$j] eq "fullName"){
		$p3 = $p2->[$j+1];
		if ($p3->[1]){
		    warn "<fullName> can only contain text";
		}
		$filetypes{$typename}->{fullName} = $p3->[2];
	    }
	    elsif ($p2->[$j] eq "ext"){
		$p3 = $p2->[$j+1]; # p3 points to contents of ext element
		if ($p3->[1]){
		    warn "<ext> can only contain text";
		    next;
		}
		push @{ $filetypes{$typename}->{extensions} }, $p3->[2];
	    }
	    elsif ($p2->[$j] eq "name"){
		$p3 = $p2->[$j+1]; # p3 points to contents of name element
		if ($p3->[1]){
		    warn "<name> can only contain text";
		    next;
		}
		push @{ $filetypes{$typename}->{names} }, $p3->[2];
	    }
	    elsif ($p2->[$j] eq "option"){	    # read options
		$p3 = $p2->[$j+1]; # p3 points to contents of name element
		if (defined $p3->[1]){
		    warn "<option> can not have anything in it";
		    next;
		}
		$p3 = $p3->[0]; # p3 now points to attribute hash of <option/>
		foreach (keys %$p3){
		    # set link option
		    if (/list/){
			if ($p3->{list} =~ /(yes|no)/){
			    $filetypes{$typename}->{options}->{list} =
				$p3->{list}
			}
			else {
			    warn "acceptable values for list option are " .
				"\"yes\" or \"no\"";
			    next;
			}
		    }
		    # set thumbnail option
		    elsif (/thumbnail/){
			if ($p3->{thumbnail} =~ /(yes|no)/){
			    $filetypes{$typename}->{options}->{thumbnail} =
				$p3->{thumbnail}
			}
			else {
			    warn "acceptable values for thumbnail option are" .
				" \"yes\" or \"no\"";
			    next;
			}
		    }
		    # set setvar option
		    elsif (/setvar/){
			$filetypes{$typename}->{options}->{setvar} =
			    $p3->{setvar}
		    }
		    # set incorporate option
		    elsif (/incorporate/){
			if ($p3->{incorporate} =~ /(yes|no)/){
			    $filetypes{$typename}->{options}->{incorporate} =
				$p3->{incorporate}
			}
			else {
			    warn "acceptable values for incorporate option ".
				"are \"yes\" or \"no\"";
			    next;
			}
		    }
		    # set caption option
		    elsif (/caption/){
			if ($p3->{caption} =~ /(yes|no)/){
			    $filetypes{$typename}->{options}->{caption} =
				$p3->{caption}
			}
			else {
			    warn "acceptable values for caption option ".
				"are \"yes\" or \"no\"";
			    next;
			}
		    }
		    # all other options invalid
		    else {
			warn "option $_ is unknown";
			next;
		    }
		}
		# end options reading
	    }
	    else {
		warn "<$p2->[$j]> is not allowed in <filetype>";
		next;
	    }
	}
    }
}

sub test_read_file_types {
    my ($i, @j);
    foreach $i (keys %filetypes){
	@j = @{ $filetypes{$i}->{extensions} };
	print "$i has " . scalar(@j) . " (aka " .
	    $filetypes{$i}->{pluralName} . ") elements:\n";
	foreach (@j){ print "   $_\n" }
    }
}

######################################################################
# subroutine read_config_file -  Read config file and set variables
#
# XML tree diagram
#
# tree
#   |
#   |- "config"
#   `-.
#     |- config attributes
#     |
#     |- "var"
#     |-.
#     | `- var attributes ("var"="val")
#       ...
#
######################################################################

sub read_config_file {

    my $parser = new XML::Parser(Style=>'Tree', ErrorContext=>2);
    my $tree = $parser->parsefile("$config_file");
    die '<config> isn\'t the top-level element'
	unless $tree->[0] eq "config";
    
    # read filetypes
    my $p = $tree->[1];  # p points to contents of config element

    # loop through contents of filetype_config
    for (my $i = 1; $i<@$p; $i+=2){

	# should not be seeing non-whitespace text, only elements
	if (not $p->[$i]){
	    warn 'text not allowed in <config>' unless $p->[$i+1] =~ /^\s*$/;
	    next;
	}
	if ($p->[$i] ne "setvar"){
	    warn "<$p->[$i]> not allowed in <config>";
	    next;
	}

	my $p2 = $p->[$i+1];  # p2 points to contents of var element
	my $p3 = $p2->[0];    # p3 points to attribute hash of var element
	
	# set variables
	my $eval_string;
	foreach (keys %$p3){
	    $eval_string = "\$$_=\"" . $p3->{$_} . "\"";
	    # print "Evaluating: $eval_string\n";
	    eval $eval_string;
	}

	# warn about elements in var
	if (@$p2>1){
	    warn "<var> should not have any contents, only atributes.";
	    print "Contents are:\n";
	    print join("\n",@{$p2});
	    print "\nEnd contents.\n";
	}
    }
}

sub test_read_config_file {
    $dir_list = "default";
    print "\$dir_list (before) = $dir_list\n";
    read_config_file();
    print "\$dir_list (after) = $dir_list\n";
    exit;
}

######################################################################
# subroutine date - Returns date string, given a time
######################################################################

sub date {
    my $time;
    if (@_){ $time = shift }
    else   { $time = time }
    my $date = "";
    my ($sec, $min, $hour, $mday, $mon, $year,
        $wday, $yday, $isdst) = localtime($time);
    my ($day, $month);
    $day = (qw(Sunday Monday Tuesday Wednesday Thursday Friday Saturday))[$wday];
    $month = (qw(January February March April May June July August September Oct
ober November December))[$mon];
    if    ($mday >= 4 &&
           $mday <= 20)  { $mday .= "th" }
    elsif ($mday%10 == 3){ $mday .= "rd" }
    elsif ($mday%10 == 2){ $mday .= "nd" }
    elsif ($mday%10 == 1){ $mday .= "st" }
    else                 { $mday .= "th" }
    if ($year<50){ $year += 2000}
    else{ $year += 1900 }
    return "$day $month $mday, $year";
}

######################################################################
# subroutine filesize - Returns small string describing file size
#     $file - the file to look at
######################################################################

sub filesize {
    my $file = shift;
    my $size = (stat $file)[7];
    if ($size < 2**10){
	return $size . "B";
    }
    elsif ($size < 2**20*10){
	return sprintf("%.0fK", $size/2**10);
    }
    else {
	return sprintf("%.1fM", $size/2**20);
    }
}

######################################################################
# subroutine makedrawfiles - Make images from drawShapes
#     $target_dir - place to put the images (0.jpg, 1.jpg, 2.jpg)
#     $seed - any string to serve as a random seed
#             if $seed is a directory with a param file, then use.
#
# 0.jpg - darkest one, the accent on the left sidebar
# 1.jpg - medium darkness, the background of the left sidebar
# 2.jpg - brightest, the background of the main page
######################################################################

sub makedrawfiles {
    my ($target_dir, $seed) = @_;

    # seed rand
    my ($i, $j);
    my $val = 1;
    my $val_max = 25371;
    foreach $i (split //, $seed){
	$j = vec($i, 0, 8);
	$val = ($j * $val) % $val_max;
    }
    srand 0;
    for ($i=0;$i<$val%362;$i++){
	rand();
    }
    
    # pick some values
    my $pi = atan2(1,0)*2.0;
    my $hue = rand() * 2.0 * $pi;
    my @values = (0.25, 0.5, 1.0);
    my $picnum;

    # create images
    my $intermediate_img;
    for ($picnum=2; $picnum>=0; $picnum--){
	$intermediate_img = "$target_dir/$picnum.ppm";

	# detect drawShapes parameter file
	if (-r "$seed/$drawshapes_param"){ # use parameter file
	    print "Using $seed/$drawshapes_param for drawShapes\n";

	    # read parameters
	    my %params;
	    open PARAM, "<$seed/$drawshapes_param" or die $!;
	    while (<PARAM>){
		chomp;
		next unless /^(\w+)\s*=\s*([.\d]+)$/;
		print "Setting $1 to $2\n";
		$params{$1} = $2;
	    }
	    $params{"Value"} = 0.8 unless $params{"Value"};
	    $params{"Value"} *= $values[$picnum];
	    my $options = "";
	    foreach (keys %params){
		$options .= "$_ $params{$_} ";
	    }
	    chop $options;
	    print "\$options = $options\n";
	    
	    open DRAW,
	    "$drawshapes_cmd $options |" or
		die "Couldn't open drawshapes command";
	}
	else {
	    print "Using default values for drawShapes\n";
	    $_ = 0.8 * $values[$picnum];
	    open DRAW,
	    "$drawshapes_cmd Watermark 0 Hue $hue Value $_ |" or
		die "Couldn't open drawshapes command";
	}
	
	print "Opening file $intermediate_img\n";
	open OUT, ">$intermediate_img" or
	    die "Couldn't open target drawshapes file $intermediate_img";
	while (<DRAW>){
	    print OUT;
	}
	close OUT;
	close DRAW;
	system "$convert_cmd", "$intermediate_img", "$target_dir/$picnum.jpg";
	if ($?){
	    warn "Couldn't convert image $intermediate_img: $?";
	}
	unlink $intermediate_img or die $!;

	# detect drawShapes error
	if (!-e "$target_dir/$picnum.jpg"){
	    # use default image
	    copy_file($target_dir, "$mirror_dir/$crawler_subdir/$picnum.jpg")
		or die("Couldn't make make background image " .
		       "$target_dir/$picnum.jpg");
	}
    }
}
